from typing import Optional
from typing import Type

from falcon.request import Request
from falcon.response import Response

from openapi_core import OpenAPI
from openapi_core.contrib.falcon.requests import FalconAsgiOpenAPIRequest
from openapi_core.contrib.falcon.requests import FalconOpenAPIRequest
from openapi_core.contrib.falcon.responses import FalconAsgiOpenAPIResponse
from openapi_core.contrib.falcon.responses import FalconOpenAPIResponse
from openapi_core.unmarshalling.processors import AsyncUnmarshallingProcessor
from openapi_core.unmarshalling.processors import UnmarshallingProcessor
from openapi_core.unmarshalling.typing import ErrorsHandlerCallable


class FalconIntegration(UnmarshallingProcessor[Request, Response]):
    request_cls = FalconOpenAPIRequest
    response_cls = FalconOpenAPIResponse

    def get_openapi_request(self, request: Request) -> FalconOpenAPIRequest:
        return self.request_cls(request)

    def get_openapi_response(
        self, response: Response
    ) -> FalconOpenAPIResponse:
        assert self.response_cls is not None
        return self.response_cls(response)

    def should_validate_response(self) -> bool:
        return self.response_cls is not None

    def handle_response(
        self,
        request: Request,
        response: Response,
        errors_handler: ErrorsHandlerCallable[Response],
    ) -> Response:
        if not self.should_validate_response():
            return response
        return super().handle_response(request, response, errors_handler)


class AsyncFalconIntegration(AsyncUnmarshallingProcessor[Request, Response]):
    request_cls: Type[FalconAsgiOpenAPIRequest] = FalconAsgiOpenAPIRequest
    response_cls: Optional[Type[FalconAsgiOpenAPIResponse]] = (
        FalconAsgiOpenAPIResponse
    )

    def __init__(
        self,
        openapi: OpenAPI,
        request_cls: Type[FalconAsgiOpenAPIRequest] = FalconAsgiOpenAPIRequest,
        response_cls: Optional[Type[FalconAsgiOpenAPIResponse]] = (
            FalconAsgiOpenAPIResponse
        ),
    ):
        super().__init__(openapi)
        self.request_cls = request_cls or self.request_cls
        self.response_cls = response_cls

    async def get_openapi_request(
        self, request: Request
    ) -> FalconAsgiOpenAPIRequest:
        return await self.request_cls.from_request(request)

    async def get_openapi_response(
        self, response: Response
    ) -> FalconAsgiOpenAPIResponse:
        assert self.response_cls is not None
        return await self.response_cls.from_response(response)

    def should_validate_response(self) -> bool:
        return self.response_cls is not None

    async def handle_response(
        self,
        request: Request,
        response: Response,
        errors_handler: ErrorsHandlerCallable[Response],
    ) -> Response:
        if not self.should_validate_response():
            return response
        return await super().handle_response(request, response, errors_handler)



import warnings
from typing import Any
from typing import Generator
from typing import Mapping
from typing import Optional
from typing import Tuple

from falcon.request import Request


def serialize_body(
    request: Request,
    media: Any,
    content_type: str,
) -> Optional[bytes]:
    """Serialize request body using media handlers."""
    handler, _, _ = request.options.media_handlers._resolve(
        content_type,
        request.options.default_media_type,
    )
    try:
        body = handler.serialize(media, content_type=content_type)
    # multipart form serialization is not supported
    except NotImplementedError:
        warnings.warn(f"body serialization for {content_type} not supported")
        return None
    assert isinstance(body, bytes)
    return body


def unpack_params(
    params: Mapping[str, Any],
) -> Generator[Tuple[str, Any], None, None]:
    for k, v in params.items():
        if isinstance(v, list):
            for v2 in v:
                yield (k, v2)
        else:
            yield (k, v)



from dataclasses import dataclass
from typing import Any

from openapi_core.deserializing.exceptions import DeserializeError


@dataclass
class CastError(DeserializeError):
    """Schema cast operation error"""

    value: Any
    type: str | None

    def __str__(self) -> str:
        return f"Failed to cast value to {self.type} type: {self.value}"



from openapi_core.contrib.falcon.middlewares import FalconASGIOpenAPIMiddleware
from openapi_core.contrib.falcon.middlewares import FalconOpenAPIMiddleware
from openapi_core.contrib.falcon.middlewares import FalconWSGIOpenAPIMiddleware
from openapi_core.contrib.falcon.requests import FalconOpenAPIRequest
from openapi_core.contrib.falcon.responses import FalconOpenAPIResponse

__all__ = [
    "FalconASGIOpenAPIMiddleware",
    "FalconOpenAPIMiddleware",
    "FalconWSGIOpenAPIMiddleware",
    "FalconOpenAPIRequest",
    "FalconOpenAPIResponse",
]



from typing import Mapping
from typing import Optional

from jsonschema_path import SchemaPath

from openapi_core.casting.schemas.factories import SchemaCastersFactory
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.deserializers import (
    MediaTypeDeserializer,
)
from openapi_core.deserializing.media_types.deserializers import (
    MediaTypesDeserializer,
)
from openapi_core.deserializing.styles.datatypes import StyleDeserializersDict
from openapi_core.deserializing.styles.factories import (
    StyleDeserializersFactory,
)
from openapi_core.validation.schemas.validators import SchemaValidator


class MediaTypeDeserializersFactory:
    def __init__(
        self,
        style_deserializers_factory: StyleDeserializersFactory,
        media_type_deserializers: Optional[MediaTypeDeserializersDict] = None,
    ):
        self.style_deserializers_factory = style_deserializers_factory
        if media_type_deserializers is None:
            media_type_deserializers = {}
        self.media_type_deserializers = media_type_deserializers

    @classmethod
    def from_schema_casters_factory(
        cls,
        schema_casters_factory: SchemaCastersFactory,
        style_deserializers: Optional[StyleDeserializersDict] = None,
        media_type_deserializers: Optional[MediaTypeDeserializersDict] = None,
    ) -> "MediaTypeDeserializersFactory":
        from openapi_core.deserializing.media_types import (
            media_type_deserializers as default_media_type_deserializers,
        )
        from openapi_core.deserializing.styles import (
            style_deserializers as default_style_deserializers,
        )

        style_deserializers_factory = StyleDeserializersFactory(
            schema_casters_factory,
            style_deserializers=style_deserializers
            or default_style_deserializers,
        )
        return cls(
            style_deserializers_factory,
            media_type_deserializers=media_type_deserializers
            or default_media_type_deserializers,
        )

    def create(
        self,
        spec: SchemaPath,
        mimetype: str,
        schema: Optional[SchemaPath] = None,
        schema_validator: Optional[SchemaValidator] = None,
        parameters: Optional[Mapping[str, str]] = None,
        encoding: Optional[SchemaPath] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
    ) -> MediaTypeDeserializer:
        if parameters is None:
            parameters = {}
        if extra_media_type_deserializers is None:
            extra_media_type_deserializers = {}
        media_types_deserializer = MediaTypesDeserializer(
            self.media_type_deserializers,
            extra_media_type_deserializers,
        )

        # Create schema caster for urlencoded/multipart content types
        # Only create if both schema and schema_validator are provided
        schema_caster = None
        if (
            schema is not None
            and schema_validator is not None
            and (
                mimetype == "application/x-www-form-urlencoded"
                or mimetype.startswith("multipart")
            )
        ):
            schema_caster = (
                self.style_deserializers_factory.schema_casters_factory.create(
                    spec, schema
                )
            )

        return MediaTypeDeserializer(
            spec,
            self.style_deserializers_factory,
            media_types_deserializer,
            mimetype,
            schema=schema,
            schema_validator=schema_validator,
            schema_caster=schema_caster,
            encoding=encoding,
            **parameters,
        )



"""OpenAPI core validation exceptions module"""

from dataclasses import dataclass
from typing import Any

from openapi_core.exceptions import OpenAPIError


def _schema_error_to_dict(schema_error: Exception) -> dict[str, Any]:
    message = getattr(schema_error, "message", str(schema_error))
    raw_path = getattr(schema_error, "path", ())
    try:
        path = list(raw_path)
    except TypeError:
        path = []
    return {
        "message": message,
        "path": path,
    }


@dataclass
class ValidationError(OpenAPIError):
    @property
    def details(self) -> dict[str, Any]:
        cause = self.__cause__
        schema_errors: list[dict[str, Any]] = []
        if cause is not None:
            cause_schema_errors = getattr(cause, "schema_errors", None)
            if cause_schema_errors is not None:
                schema_errors = [
                    _schema_error_to_dict(schema_error)
                    for schema_error in cause_schema_errors
                ]

        return {
            "message": str(self),
            "error_type": self.__class__.__name__,
            "cause_type": (
                cause.__class__.__name__ if cause is not None else None
            ),
            "schema_errors": schema_errors,
        }

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.__cause__}"



"""OpenAPI core validation request datatypes module"""

from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Mapping
from typing import Union

from werkzeug.datastructures import Headers
from werkzeug.datastructures import ImmutableMultiDict

# Type alias for headers that accepts both Mapping and werkzeug Headers
HeadersType = Union[Mapping[str, Any], Headers]


@dataclass
class RequestParameters:
    """OpenAPI request parameters dataclass.

    Attributes:
        query
            Query string parameters as MultiDict. Must support getlist method.
        header
            Request headers as Headers.
        cookie
            Request cookies as MultiDict.
        path
            Path parameters as dict. Gets resolved against spec if empty.
    """

    query: Mapping[str, Any] = field(default_factory=ImmutableMultiDict)
    header: HeadersType = field(default_factory=Headers)
    cookie: Mapping[str, Any] = field(default_factory=ImmutableMultiDict)
    path: Mapping[str, Any] = field(default_factory=dict)

    def __getitem__(self, location: str) -> Any:
        return getattr(self, location)


@dataclass
class Parameters:
    query: Mapping[str, Any] = field(default_factory=dict)
    header: Mapping[str, Any] = field(default_factory=dict)
    cookie: Mapping[str, Any] = field(default_factory=dict)
    path: Mapping[str, Any] = field(default_factory=dict)



"""OpenAPI core contrib falcon middlewares module"""

from typing import Any
from typing import Optional
from typing import Type
from typing import cast

from falcon.request import Request
from falcon.response import Response
from jsonschema_path import SchemaPath

from openapi_core import OpenAPI
from openapi_core.contrib.falcon.handlers import FalconOpenAPIErrorsHandler
from openapi_core.contrib.falcon.handlers import (
    FalconOpenAPIValidRequestHandler,
)
from openapi_core.contrib.falcon.integrations import AsyncFalconIntegration
from openapi_core.contrib.falcon.integrations import FalconIntegration
from openapi_core.contrib.falcon.requests import FalconAsgiOpenAPIRequest
from openapi_core.contrib.falcon.requests import FalconOpenAPIRequest
from openapi_core.contrib.falcon.responses import FalconAsgiOpenAPIResponse
from openapi_core.contrib.falcon.responses import FalconOpenAPIResponse
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.request.types import RequestUnmarshallerType
from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType

_DEFAULT_ASYNC = object()


class FalconWSGIOpenAPIMiddleware(FalconIntegration):
    """OpenAPI middleware for Falcon WSGI applications.

    This class wires Falcon's synchronous middleware hooks to the
    synchronous OpenAPI integration.
    """

    valid_request_handler_cls = FalconOpenAPIValidRequestHandler
    errors_handler_cls: Type[FalconOpenAPIErrorsHandler] = (
        FalconOpenAPIErrorsHandler
    )

    def __init__(
        self,
        openapi: OpenAPI,
        request_cls: Type[FalconOpenAPIRequest] = FalconOpenAPIRequest,
        response_cls: Type[FalconOpenAPIResponse] = FalconOpenAPIResponse,
        errors_handler_cls: Type[
            FalconOpenAPIErrorsHandler
        ] = FalconOpenAPIErrorsHandler,
        **kwargs: Any,
    ):
        super().__init__(openapi)
        self.request_cls = request_cls or self.request_cls
        self.response_cls = response_cls or self.response_cls
        self.errors_handler_cls = errors_handler_cls or self.errors_handler_cls

    @classmethod
    def from_spec(
        cls,
        spec: SchemaPath,
        request_unmarshaller_cls: Optional[RequestUnmarshallerType] = None,
        response_unmarshaller_cls: Optional[ResponseUnmarshallerType] = None,
        request_cls: Type[FalconOpenAPIRequest] = FalconOpenAPIRequest,
        response_cls: Type[FalconOpenAPIResponse] = FalconOpenAPIResponse,
        errors_handler_cls: Type[
            FalconOpenAPIErrorsHandler
        ] = FalconOpenAPIErrorsHandler,
        **kwargs: Any,
    ) -> "FalconWSGIOpenAPIMiddleware":
        openapi = OpenAPI.build(
            spec,
            request_unmarshaller_cls=request_unmarshaller_cls,
            response_unmarshaller_cls=response_unmarshaller_cls,
        )
        return cls(
            openapi,
            request_cls=request_cls,
            response_cls=response_cls,
            errors_handler_cls=errors_handler_cls,
            **kwargs,
        )

    def process_request(self, req: Request, resp: Response) -> None:
        valid_handler = self.valid_request_handler_cls(req, resp)
        errors_handler = self.errors_handler_cls(req, resp)
        self.handle_request(req, valid_handler, errors_handler)

    def process_response(
        self, req: Request, resp: Response, resource: Any, req_succeeded: bool
    ) -> None:
        errors_handler = self.errors_handler_cls(req, resp)
        self.handle_response(req, resp, errors_handler)


class FalconASGIOpenAPIMiddleware(AsyncFalconIntegration):
    """OpenAPI middleware for Falcon ASGI applications.

    This class wires Falcon's asynchronous middleware hooks to the
    asynchronous OpenAPI integration.
    """

    valid_request_handler_cls = FalconOpenAPIValidRequestHandler
    errors_handler_cls: Type[FalconOpenAPIErrorsHandler] = (
        FalconOpenAPIErrorsHandler
    )

    def __init__(
        self,
        openapi: OpenAPI,
        request_cls: Type[FalconAsgiOpenAPIRequest] = FalconAsgiOpenAPIRequest,
        response_cls: Optional[Type[FalconAsgiOpenAPIResponse]] = (
            FalconAsgiOpenAPIResponse
        ),
        errors_handler_cls: Type[
            FalconOpenAPIErrorsHandler
        ] = FalconOpenAPIErrorsHandler,
        **kwargs: Any,
    ):
        super().__init__(
            openapi,
            request_cls=request_cls,
            response_cls=response_cls,
        )
        self.errors_handler_cls = errors_handler_cls or self.errors_handler_cls

    @classmethod
    def from_spec(
        cls,
        spec: SchemaPath,
        request_unmarshaller_cls: Optional[RequestUnmarshallerType] = None,
        response_unmarshaller_cls: Optional[ResponseUnmarshallerType] = None,
        request_cls: Type[FalconAsgiOpenAPIRequest] = FalconAsgiOpenAPIRequest,
        response_cls: Optional[Type[FalconAsgiOpenAPIResponse]] = (
            FalconAsgiOpenAPIResponse
        ),
        errors_handler_cls: Type[
            FalconOpenAPIErrorsHandler
        ] = FalconOpenAPIErrorsHandler,
        **kwargs: Any,
    ) -> "FalconASGIOpenAPIMiddleware":
        openapi = OpenAPI.build(
            spec,
            request_unmarshaller_cls=request_unmarshaller_cls,
            response_unmarshaller_cls=response_unmarshaller_cls,
        )
        return cls(
            openapi,
            request_cls=request_cls,
            response_cls=response_cls,
            errors_handler_cls=errors_handler_cls,
            **kwargs,
        )

    async def process_request_async(
        self, req: Request, resp: Response
    ) -> None:
        errors_handler = self.errors_handler_cls(req, resp)
        valid_request_handler = self.valid_request_handler_cls(req, resp)

        async def async_valid_request_handler(
            request_unmarshal_result: RequestUnmarshalResult,
        ) -> Response:
            return valid_request_handler(request_unmarshal_result)

        await self.handle_request(
            req,
            async_valid_request_handler,
            errors_handler,
        )

    async def process_response_async(
        self,
        req: Request,
        resp: Response,
        resource: Any,
        req_succeeded: bool,
    ) -> None:
        errors_handler = self.errors_handler_cls(req, resp)
        await self.handle_response(req, resp, errors_handler)


class FalconOpenAPIMiddleware:
    """OpenAPI middleware compatible with both WSGI and ASGI Falcon apps.

    This class delegates to transport-specific middleware implementations:
    :class:`FalconWSGIOpenAPIMiddleware` for sync hooks and
    :class:`FalconASGIOpenAPIMiddleware` for async hooks.
    """

    def __init__(
        self,
        openapi: OpenAPI,
        request_cls: Type[FalconOpenAPIRequest] = FalconOpenAPIRequest,
        response_cls: Type[FalconOpenAPIResponse] = FalconOpenAPIResponse,
        request_async_cls: Any = _DEFAULT_ASYNC,
        response_async_cls: Any = _DEFAULT_ASYNC,
        errors_handler_cls: Type[
            FalconOpenAPIErrorsHandler
        ] = FalconOpenAPIErrorsHandler,
        **kwargs: Any,
    ):
        if request_async_cls is _DEFAULT_ASYNC:
            request_async_cls = FalconAsgiOpenAPIRequest
        if response_async_cls is _DEFAULT_ASYNC:
            response_async_cls = (
                FalconAsgiOpenAPIResponse if response_cls is not None else None
            )

        self.wsgi_middleware = FalconWSGIOpenAPIMiddleware(
            openapi,
            request_cls=request_cls,
            response_cls=response_cls,
            errors_handler_cls=errors_handler_cls,
            **kwargs,
        )
        self.asgi_middleware = FalconASGIOpenAPIMiddleware(
            openapi,
            request_cls=cast(
                Type[FalconAsgiOpenAPIRequest], request_async_cls
            ),
            response_cls=cast(
                Optional[Type[FalconAsgiOpenAPIResponse]],
                response_async_cls,
            ),
            errors_handler_cls=errors_handler_cls,
            **kwargs,
        )

    @classmethod
    def from_spec(
        cls,
        spec: SchemaPath,
        request_unmarshaller_cls: Optional[RequestUnmarshallerType] = None,
        response_unmarshaller_cls: Optional[ResponseUnmarshallerType] = None,
        request_cls: Type[FalconOpenAPIRequest] = FalconOpenAPIRequest,
        response_cls: Type[FalconOpenAPIResponse] = FalconOpenAPIResponse,
        request_async_cls: Any = _DEFAULT_ASYNC,
        response_async_cls: Any = _DEFAULT_ASYNC,
        errors_handler_cls: Type[
            FalconOpenAPIErrorsHandler
        ] = FalconOpenAPIErrorsHandler,
        **kwargs: Any,
    ) -> "FalconOpenAPIMiddleware":
        openapi = OpenAPI.build(
            spec,
            request_unmarshaller_cls=request_unmarshaller_cls,
            response_unmarshaller_cls=response_unmarshaller_cls,
        )
        return cls(
            openapi,
            request_cls=request_cls,
            response_cls=response_cls,
            request_async_cls=request_async_cls,
            response_async_cls=response_async_cls,
            errors_handler_cls=errors_handler_cls,
            **kwargs,
        )

    def process_request(self, req: Request, resp: Response) -> None:
        self.wsgi_middleware.process_request(req, resp)

    def process_response(
        self, req: Request, resp: Response, resource: Any, req_succeeded: bool
    ) -> None:
        self.wsgi_middleware.process_response(
            req,
            resp,
            resource,
            req_succeeded,
        )

    async def process_request_async(
        self, req: Request, resp: Response
    ) -> None:
        await self.asgi_middleware.process_request_async(req, resp)

    async def process_response_async(
        self,
        req: Request,
        resp: Response,
        resource: Any,
        req_succeeded: bool,
    ) -> None:
        await self.asgi_middleware.process_response_async(
            req,
            resp,
            resource,
            req_succeeded,
        )



from typing import TYPE_CHECKING
from typing import Any
from typing import Mapping
from typing import Optional
from xml.etree.ElementTree import ParseError

from jsonschema_path import SchemaPath

from openapi_core.deserializing.media_types.datatypes import (
    DeserializerCallable,
)
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.exceptions import (
    MediaTypeDeserializeError,
)
from openapi_core.deserializing.styles.factories import (
    StyleDeserializersFactory,
)
from openapi_core.schema.encodings import get_content_type
from openapi_core.schema.parameters import get_style_and_explode
from openapi_core.schema.protocols import SuportsGetAll
from openapi_core.schema.protocols import SuportsGetList
from openapi_core.schema.schemas import get_properties
from openapi_core.validation.schemas.validators import SchemaValidator

if TYPE_CHECKING:
    from openapi_core.casting.schemas.casters import SchemaCaster


class MediaTypesDeserializer:
    def __init__(
        self,
        media_type_deserializers: Optional[MediaTypeDeserializersDict] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
    ):
        if media_type_deserializers is None:
            media_type_deserializers = {}
        self.media_type_deserializers = media_type_deserializers
        if extra_media_type_deserializers is None:
            extra_media_type_deserializers = {}
        self.extra_media_type_deserializers = extra_media_type_deserializers

    def deserialize(
        self, mimetype: str, value: bytes, **parameters: str
    ) -> Any:
        deserializer_callable = self.get_deserializer_callable(mimetype)

        try:
            return deserializer_callable(value, **parameters)
        except (ParseError, ValueError, TypeError, AttributeError):
            raise MediaTypeDeserializeError(mimetype, value)

    def get_deserializer_callable(
        self,
        mimetype: str,
    ) -> DeserializerCallable:
        if mimetype in self.extra_media_type_deserializers:
            return self.extra_media_type_deserializers[mimetype]
        return self.media_type_deserializers[mimetype]


class MediaTypeDeserializer:
    def __init__(
        self,
        spec: SchemaPath,
        style_deserializers_factory: StyleDeserializersFactory,
        media_types_deserializer: MediaTypesDeserializer,
        mimetype: str,
        schema: Optional[SchemaPath] = None,
        schema_validator: Optional[SchemaValidator] = None,
        schema_caster: Optional["SchemaCaster"] = None,
        encoding: Optional[SchemaPath] = None,
        **parameters: str,
    ):
        self.spec = spec
        self.style_deserializers_factory = style_deserializers_factory
        self.media_types_deserializer = media_types_deserializer
        self.mimetype = mimetype
        self.schema = schema
        self.schema_validator = schema_validator
        self.schema_caster = schema_caster
        self.encoding = encoding
        self.parameters = parameters

    def deserialize(self, value: bytes) -> Any:
        deserialized = self.media_types_deserializer.deserialize(
            self.mimetype, value, **self.parameters
        )

        if (
            self.mimetype != "application/x-www-form-urlencoded"
            and not self.mimetype.startswith("multipart")
        ):
            return deserialized

        # decode multipart request bodies if schema provided
        if self.schema is not None:
            return self.decode(deserialized)

        return deserialized

    def evolve(
        self,
        schema: SchemaPath,
        mimetype: Optional[str] = None,
    ) -> "MediaTypeDeserializer":
        cls = self.__class__

        schema_validator = None
        if self.schema_validator is not None:
            schema_validator = self.schema_validator.evolve(schema)

        schema_caster = None
        if self.schema_caster is not None:
            schema_caster = self.schema_caster.evolve(schema)

        return cls(
            self.spec,
            self.style_deserializers_factory,
            self.media_types_deserializer,
            mimetype=mimetype or self.mimetype,
            schema=schema,
            schema_validator=schema_validator,
            schema_caster=schema_caster,
        )

    def decode(
        self, location: Mapping[str, Any], schema_only: bool = False
    ) -> Mapping[str, Any]:
        # schema is required for multipart
        assert self.schema is not None
        properties: dict[str, Any] = {}

        # For urlencoded/multipart, use caster for oneOf/anyOf detection if validator available
        if self.schema_validator is not None:
            one_of_schema = self.schema_validator.get_one_of_schema(
                location, caster=self.schema_caster
            )
            if one_of_schema is not None:
                one_of_properties = self.evolve(one_of_schema).decode(
                    location, schema_only=True
                )
                properties.update(one_of_properties)

            any_of_schemas = self.schema_validator.iter_any_of_schemas(
                location, caster=self.schema_caster
            )
            for any_of_schema in any_of_schemas:
                any_of_properties = self.evolve(any_of_schema).decode(
                    location, schema_only=True
                )
                properties.update(any_of_properties)

            all_of_schemas = self.schema_validator.iter_all_of_schemas(
                location
            )
            for all_of_schema in all_of_schemas:
                all_of_properties = self.evolve(all_of_schema).decode(
                    location, schema_only=True
                )
                properties.update(all_of_properties)

        for prop_name, prop_schema in get_properties(self.schema).items():
            try:
                properties[prop_name] = self.decode_property(
                    prop_name, prop_schema, location
                )
            except KeyError:
                if "default" not in prop_schema:
                    continue
                properties[prop_name] = (prop_schema / "default").read_value()

        if schema_only:
            return properties

        return properties

    def decode_property(
        self,
        prop_name: str,
        prop_schema: SchemaPath,
        location: Mapping[str, Any],
    ) -> Any:
        if self.encoding is None or prop_name not in self.encoding:
            if self.mimetype == "application/x-www-form-urlencoded":
                # default serialization strategy for complex objects
                # in the application/x-www-form-urlencoded
                return self.decode_property_style(
                    prop_name,
                    prop_schema,
                    location,
                    SchemaPath.from_dict({"style": "form"}),
                )
            return self.decode_property_content_type(
                prop_name, prop_schema, location
            )

        prep_encoding = self.encoding / prop_name
        if (
            "style" not in prep_encoding
            and "explode" not in prep_encoding
            and "allowReserved" not in prep_encoding
        ):
            return self.decode_property_content_type(
                prop_name, prop_schema, location, prep_encoding
            )

        return self.decode_property_style(
            prop_name, prop_schema, location, prep_encoding
        )

    def decode_property_style(
        self,
        prop_name: str,
        prop_schema: SchemaPath,
        location: Mapping[str, Any],
        prep_encoding: SchemaPath,
    ) -> Any:
        prop_style, prop_explode = get_style_and_explode(
            prep_encoding, default_location="query"
        )
        prop_deserializer = self.style_deserializers_factory.create(
            self.spec, prop_schema, prop_style, prop_explode, name=prop_name
        )
        return prop_deserializer.deserialize(location)

    def decode_property_content_type(
        self,
        prop_name: str,
        prop_schema: SchemaPath,
        location: Mapping[str, Any],
        prop_encoding: Optional[SchemaPath] = None,
    ) -> Any:
        prop_content_type = get_content_type(prop_schema, prop_encoding)
        prop_deserializer = self.evolve(
            prop_schema,
            mimetype=prop_content_type,
        )
        prop_schema_type = (prop_schema / "type").read_str("")
        if (
            self.mimetype.startswith("multipart")
            and prop_schema_type == "array"
        ):
            if isinstance(location, SuportsGetAll):
                value = location.getall(prop_name)
                return list(map(prop_deserializer.deserialize, value))
            if isinstance(location, SuportsGetList):
                value = location.getlist(prop_name)
                return list(map(prop_deserializer.deserialize, value))

        return prop_deserializer.deserialize(location[prop_name])



import re
from dataclasses import dataclass


@dataclass(frozen=True)
class PathMatchResult:
    """Result of path parsing."""

    named: dict[str, str]


class PathParser:
    """Parses path patterns with parameters into regex and matches against URLs."""

    _PARAM_PATTERN = r"[^/]*"

    def __init__(
        self, pattern: str, pre_expression: str = "", post_expression: str = ""
    ) -> None:
        self.pattern = pattern
        self._group_to_name: dict[str, str] = {}

        regex_body = self._compile_template_to_regex(pattern)
        self._expression = f"{pre_expression}{regex_body}{post_expression}"
        self._compiled = re.compile(self._expression)

    def search(self, text: str) -> PathMatchResult | None:
        """Searches for a match in the given text."""
        match = self._compiled.search(text)
        return self._to_result(match)

    def parse(self, text: str) -> PathMatchResult | None:
        """Parses the entire text for a match."""
        match = self._compiled.fullmatch(text)
        return self._to_result(match)

    def _compile_template_to_regex(self, template: str) -> str:
        parts: list[str] = []
        i = 0
        group_index = 0
        while i < len(template):
            start = template.find("{", i)
            if start == -1:
                parts.append(re.escape(template[i:]))
                break
            end = template.find("}", start + 1)
            if end == -1:
                raise ValueError(f"Unmatched '{{' in template: {template!r}")

            parts.append(re.escape(template[i:start]))
            param_name = template[start + 1 : end]
            group_name = f"g{group_index}"
            group_index += 1
            self._group_to_name[group_name] = param_name
            parts.append(f"(?P<{group_name}>{self._PARAM_PATTERN})")
            i = end + 1

        return "".join(parts)

    def _to_result(
        self, match: re.Match[str] | None
    ) -> PathMatchResult | None:
        if match is None:
            return None
        return PathMatchResult(
            named={
                param_name: match.group(group_name)
                for group_name, param_name in self._group_to_name.items()
            },
        )



"""OpenAPI core contrib falcon responses module"""

from json import dumps
from typing import Any
from typing import Dict
from typing import Optional
from typing import cast

from falcon.request import Request
from falcon.request import RequestOptions
from werkzeug.datastructures import Headers
from werkzeug.datastructures import ImmutableMultiDict

from openapi_core.contrib.falcon.util import serialize_body
from openapi_core.contrib.falcon.util import unpack_params
from openapi_core.datatypes import RequestParameters

_BODY_NOT_SET = object()


class FalconOpenAPIRequest:
    def __init__(
        self,
        request: Request,
        default_when_empty: Optional[Dict[Any, Any]] = None,
    ):
        if not isinstance(request, Request):
            raise TypeError(f"'request' argument is not type of {Request}")
        self.request = request
        if default_when_empty is None:
            default_when_empty = {}
        self.default_when_empty = default_when_empty
        self._body: Any = _BODY_NOT_SET

        # Path gets deduced by path finder against spec
        self.parameters = RequestParameters(
            query=ImmutableMultiDict(unpack_params(self.request.params)),
            header=Headers(self.request.headers),
            cookie=self.request.cookies,
        )

    @property
    def host_url(self) -> str:
        assert isinstance(self.request.prefix, str)
        return self.request.prefix

    @property
    def path(self) -> str:
        assert isinstance(self.request.path, str)
        return self.request.path

    @property
    def method(self) -> str:
        assert isinstance(self.request.method, str)
        return self.request.method.lower()

    @property
    def body(self) -> Optional[bytes]:
        if self._body is not _BODY_NOT_SET:
            return cast(Optional[bytes], self._body)

        # Falcon doesn't store raw request stream.
        # That's why we need to revert deserialized data

        # Support falcon-jsonify.
        request_json = getattr(cast(Any, self.request), "json", None)
        if request_json is not None:
            self._body = dumps(request_json).encode("utf-8")
            return cast(Optional[bytes], self._body)

        media = self.request.get_media(
            default_when_empty=self.default_when_empty,
        )
        self._body = serialize_body(self.request, media, self.content_type)
        return cast(Optional[bytes], self._body)

    @property
    def content_type(self) -> str:
        if self.request.content_type:
            assert isinstance(self.request.content_type, str)
            return self.request.content_type

        assert isinstance(self.request.options, RequestOptions)
        assert isinstance(self.request.options.default_media_type, str)
        return self.request.options.default_media_type


class FalconAsgiOpenAPIRequest(FalconOpenAPIRequest):
    @classmethod
    async def from_request(
        cls,
        request: Request,
        default_when_empty: Optional[Dict[Any, Any]] = None,
    ) -> "FalconAsgiOpenAPIRequest":
        instance = cls(
            request,
            default_when_empty=default_when_empty,
        )
        media = await request.get_media(
            default_when_empty=instance.default_when_empty
        )
        instance._body = serialize_body(request, media, instance.content_type)
        return instance



"""OpenAPI core contrib falcon responses module"""

import inspect
from io import BytesIO
from itertools import tee
from typing import Any
from typing import Iterable
from typing import List

from falcon.response import Response
from werkzeug.datastructures import Headers


class FalconOpenAPIResponse:
    def __init__(self, response: Response):
        if not isinstance(response, Response):
            raise TypeError(f"'response' argument is not type of {Response}")
        self.response = response

    @property
    def data(self) -> bytes:
        if self.response.text is None:
            if self.response.stream is None:
                return b""
            if isinstance(self.response.stream, Iterable):
                resp_iter1, resp_iter2 = tee(self.response.stream)
                self.response.stream = resp_iter1
                content = b"".join(resp_iter2)
                return content
            # checks ReadableIO protocol
            if hasattr(self.response.stream, "read"):
                data = self.response.stream.read()
                self.response.stream = BytesIO(data)
                return data
        assert isinstance(self.response.text, str)
        return self.response.text.encode("utf-8")

    @property
    def status_code(self) -> int:
        return self.response.status_code

    @property
    def content_type(self) -> str:
        content_type = ""
        if self.response.content_type:
            content_type = self.response.content_type
        else:
            content_type = self.response.options.default_media_type
        return content_type

    @property
    def headers(self) -> Headers:
        return Headers(self.response.headers)


class FalconAsgiOpenAPIResponse(FalconOpenAPIResponse):
    def __init__(self, response: Response, data: bytes):
        super().__init__(response)
        self._data = data

    @classmethod
    async def from_response(
        cls,
        response: Any,
    ) -> "FalconAsgiOpenAPIResponse":
        data = await cls._get_asgi_response_data(response)
        return cls(response, data=data)

    @classmethod
    async def _get_asgi_response_data(cls, response: Any) -> bytes:
        response_any = response
        stream = response_any.stream
        if stream is None:
            data = await response_any.render_body()
            if data is None:
                return b""
            assert isinstance(data, bytes)
            return data

        charset = getattr(response_any, "charset", None) or "utf-8"
        chunks: List[bytes] = []
        stream_any = stream

        if hasattr(stream_any, "__aiter__"):
            async for chunk in stream_any:
                if chunk is None:
                    break
                if not isinstance(chunk, bytes):
                    chunk = chunk.encode(charset)
                chunks.append(chunk)
        elif hasattr(stream_any, "read"):
            while True:
                chunk = stream_any.read()
                if inspect.isawaitable(chunk):
                    chunk = await chunk
                if not chunk:
                    break
                if not isinstance(chunk, bytes):
                    chunk = chunk.encode(charset)
                chunks.append(chunk)
        elif isinstance(stream_any, Iterable):
            response_iter1, response_iter2 = tee(stream_any)
            response_any.stream = response_iter1
            for chunk in response_iter2:
                if not isinstance(chunk, bytes):
                    chunk = chunk.encode(charset)
                chunks.append(chunk)
            return b"".join(chunks)

        response_any.stream = _AsyncChunksIterator(chunks)
        return b"".join(chunks)

    @property
    def data(self) -> bytes:
        return self._data


class _AsyncChunksIterator:
    def __init__(self, chunks: List[bytes]):
        self._chunks = chunks
        self._index = 0

    def __aiter__(self) -> "_AsyncChunksIterator":
        return self

    async def __anext__(self) -> bytes:
        if self._index >= len(self._chunks):
            raise StopAsyncIteration

        chunk = self._chunks[self._index]
        self._index += 1
        return chunk



from typing import Any
from typing import Iterator
from typing import Mapping
from typing import cast

from jsonschema._utils import extras_msg
from jsonschema._utils import find_additional_properties
from jsonschema.exceptions import ValidationError
from jsonschema.protocols import Validator
from jsonschema.validators import extend


def build_forbid_unspecified_additional_properties_validator(
    validator_class: type[Validator],
) -> type[Validator]:
    properties_validator = validator_class.VALIDATORS.get("properties")
    type_validator = validator_class.VALIDATORS.get("type")

    def strict_properties(
        validator: Any,
        properties: Any,
        instance: Any,
        schema: Mapping[str, Any],
    ) -> Iterator[Any]:
        if properties_validator is not None:
            yield from properties_validator(
                validator, properties, instance, schema
            )
        yield from iter_missing_additional_properties_errors(
            validator, instance, schema
        )

    def strict_type(
        validator: Any,
        data_type: Any,
        instance: Any,
        schema: Mapping[str, Any],
    ) -> Iterator[Any]:
        if type_validator is not None:
            yield from type_validator(validator, data_type, instance, schema)

        schema_types = data_type
        if isinstance(schema_types, str):
            schema_types = [schema_types]
        if not isinstance(schema_types, list):
            return
        if "object" not in schema_types:
            return
        if "additionalProperties" in schema or "properties" in schema:
            return

        yield from iter_missing_additional_properties_errors(
            validator, instance, schema
        )

    return cast(
        type[Validator],
        extend(
            validator_class,
            validators={
                "properties": strict_properties,
                "type": strict_type,
            },
        ),
    )


def iter_missing_additional_properties_errors(
    validator: Any,
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if not validator.is_type(instance, "object"):
        return

    if "additionalProperties" in schema:
        return

    extras = sorted(set(find_additional_properties(instance, schema)))

    if extras:
        error = "Additional properties are not allowed (%s %s unexpected)"
        yield ValidationError(error % extras_msg(extras))


def build_enforce_properties_required_validator(
    validator_class: type[Validator],
) -> type[Validator]:
    properties_validator = validator_class.VALIDATORS.get("properties")

    def enforce_properties(
        validator: Any,
        properties: Any,
        instance: Any,
        schema: Mapping[str, Any],
    ) -> Iterator[Any]:
        if properties_validator is not None:
            yield from properties_validator(
                validator, properties, instance, schema
            )

        if not validator.is_type(instance, "object"):
            return

        for prop_name, prop_schema in properties.items():
            if prop_name not in instance:
                if (
                    isinstance(prop_schema, dict)
                    and prop_schema.get("writeOnly") is True
                ):
                    continue
                yield ValidationError(f"'{prop_name}' is a required property")

    return cast(
        type[Validator],
        extend(
            validator_class,
            validators={
                "properties": enforce_properties,
            },
        ),
    )



from openapi_schema_validator import OAS31_BASE_DIALECT_ID
from openapi_schema_validator import OAS32_BASE_DIALECT_ID
from openapi_schema_validator import OAS30ReadValidator
from openapi_schema_validator import OAS30WriteValidator
from openapi_schema_validator import OAS31Validator
from openapi_schema_validator import OAS32Validator

from openapi_core.validation.schemas.factories import (
    DialectSchemaValidatorsFactory,
)
from openapi_core.validation.schemas.factories import SchemaValidatorsFactory

__all__ = [
    "oas30_write_schema_validators_factory",
    "oas30_read_schema_validators_factory",
    "oas31_schema_validators_factory",
    "oas32_schema_validators_factory",
]

oas30_write_schema_validators_factory = SchemaValidatorsFactory(
    OAS30WriteValidator,
)

oas30_read_schema_validators_factory = SchemaValidatorsFactory(
    OAS30ReadValidator,
)

oas31_schema_validators_factory = DialectSchemaValidatorsFactory(
    OAS31Validator,
    OAS31_BASE_DIALECT_ID,
    # NOTE: Intentionally use OAS 3.0 format checker for OAS 3.1 to preserve
    # backward compatibility for `byte`/`binary` formats.
    # See https://github.com/python-openapi/openapi-core/issues/506
    format_checker=OAS30ReadValidator.FORMAT_CHECKER,
)

oas32_schema_validators_factory = DialectSchemaValidatorsFactory(
    OAS32Validator,
    OAS32_BASE_DIALECT_ID,
)



from typing import Any

from jsonschema_path import SchemaPath

from openapi_core.util import forcebool


def cast_primitive(value: Any, schema: SchemaPath) -> Any:
    """Cast a primitive value based on schema type."""
    schema_type = (schema / "type").read_str("")

    if schema_type == "integer":
        return int(value)
    elif schema_type == "number":
        return float(value)
    elif schema_type == "boolean":
        return forcebool(value)

    return value


def cast_value(value: Any, schema: SchemaPath, cast: bool) -> Any:
    """Recursively cast a value based on schema."""
    if not cast:
        return value

    schema_type = (schema / "type").read_str("")

    # Handle arrays
    if schema_type == "array":
        if not isinstance(value, list):
            raise ValueError(
                f"Expected list for array type, got {type(value)}"
            )
        items_schema = schema.get("items", SchemaPath.from_dict({}))
        return [cast_value(item, items_schema, cast) for item in value]

    # Handle objects
    if schema_type == "object":
        if not isinstance(value, dict):
            raise ValueError(
                f"Expected dict for object type, got {type(value)}"
            )
        properties = schema.get("properties", SchemaPath.from_dict({}))
        result = {}
        for key, val in value.items():
            if key in properties:
                prop_schema = schema / "properties" / key
                result[key] = cast_value(val, prop_schema, cast)
            else:
                result[key] = val
        return result

    # Handle primitives
    return cast_primitive(value, schema)



from copy import deepcopy
from functools import lru_cache
from typing import Any
from typing import Optional
from typing import cast

from jsonschema._format import FormatChecker
from jsonschema.protocols import Validator
from jsonschema.validators import validator_for
from jsonschema_path import SchemaPath

from openapi_core.validation.schemas._validators import (
    build_enforce_properties_required_validator,
)
from openapi_core.validation.schemas._validators import (
    build_forbid_unspecified_additional_properties_validator,
)
from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.validators import SchemaValidator


class SchemaValidatorsFactory:
    def __init__(
        self,
        schema_validator_cls: type[Validator],
        format_checker: Optional[FormatChecker] = None,
    ):
        self.schema_validator_cls = schema_validator_cls
        if format_checker is None:
            format_checker = self.schema_validator_cls.FORMAT_CHECKER
        assert format_checker is not None
        self.format_checker = format_checker

    def get_validator_cls(
        self, spec: SchemaPath, schema: SchemaPath
    ) -> type[Validator]:
        return self.schema_validator_cls

    def get_format_checker(
        self,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
    ) -> FormatChecker:
        format_checker: FormatChecker
        if format_validators is None:
            format_checker = deepcopy(cast(FormatChecker, self.format_checker))
        else:
            format_checker = FormatChecker([])
            format_checker = self._add_validators(
                cast(FormatChecker, format_checker), format_validators
            )
        format_checker = self._add_validators(
            cast(FormatChecker, format_checker), extra_format_validators
        )
        return format_checker

    def _add_validators(
        self,
        base_format_checker: FormatChecker,
        format_validators: Optional[FormatValidatorsDict] = None,
    ) -> FormatChecker:
        if format_validators is not None:
            for name, check in format_validators.items():
                base_format_checker.checks(name)(check)
        return base_format_checker

    def create(
        self,
        spec: SchemaPath,
        schema: SchemaPath,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
        forbid_unspecified_additional_properties: bool = False,
        enforce_properties_required: bool = False,
    ) -> SchemaValidator:
        validator_cls: type[Validator] = self.get_validator_cls(spec, schema)
        if enforce_properties_required:
            validator_cls = build_enforce_properties_required_validator(
                validator_cls
            )
        if forbid_unspecified_additional_properties:
            validator_cls = (
                build_forbid_unspecified_additional_properties_validator(
                    validator_cls
                )
            )

        format_checker = self.get_format_checker(
            format_validators, extra_format_validators
        )
        with schema.resolve() as resolved:
            jsonschema_validator = validator_cls(
                resolved.contents,
                _resolver=resolved.resolver,
                format_checker=format_checker,
            )

        return SchemaValidator(schema, jsonschema_validator)


class DialectSchemaValidatorsFactory(SchemaValidatorsFactory):
    def __init__(
        self,
        schema_validator_cls: type[Validator],
        default_jsonschema_dialect_id: str,
        format_checker: Optional[FormatChecker] = None,
    ):
        super().__init__(schema_validator_cls, format_checker)
        self.default_jsonschema_dialect_id = default_jsonschema_dialect_id

    def get_validator_cls(
        self, spec: SchemaPath, schema: SchemaPath
    ) -> type[Validator]:
        dialect_id = self._get_dialect_id(spec, schema)

        validator_cls = self._get_validator_class_for_dialect(dialect_id)
        if validator_cls is None:
            raise ValueError(f"Unknown JSON Schema dialect: {dialect_id!r}")

        return validator_cls

    def _get_dialect_id(
        self,
        spec: SchemaPath,
        schema: SchemaPath,
    ) -> str:
        try:
            return (schema / "$schema").read_str()
        except KeyError:
            return self._get_default_jsonschema_dialect_id(spec)

    def _get_default_jsonschema_dialect_id(self, spec: SchemaPath) -> str:
        return (spec / "jsonSchemaDialect").read_str(
            default=self.default_jsonschema_dialect_id
        )

    @lru_cache
    def _get_validator_class_for_dialect(
        self, dialect_id: str
    ) -> type[Validator] | None:
        return cast(
            type[Validator] | None,
            validator_for(
                {"$schema": dialect_id},
                default=cast(Any, None),
            ),
        )



from typing import Any
from typing import Generic
from typing import Iterable
from typing import Mapping
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

from jsonschema_path import SchemaPath

from openapi_core.casting.schemas.exceptions import CastError
from openapi_core.schema.schemas import get_properties
from openapi_core.util import BOOLEAN_FALSE_VALUES
from openapi_core.util import BOOLEAN_TRUE_VALUES
from openapi_core.util import forcebool
from openapi_core.validation.schemas.validators import SchemaValidator


class PrimitiveCaster:
    def __init__(
        self,
        schema: SchemaPath,
        schema_validator: SchemaValidator,
        schema_caster: "SchemaCaster",
    ):
        self.schema = schema
        self.schema_validator = schema_validator
        self.schema_caster = schema_caster

    def __call__(self, value: Any) -> Any:
        self.validate(value)

        return self.cast(value)

    def validate(self, value: Any) -> None:
        pass

    def cast(self, value: Any) -> Any:
        return value


class AnyCaster(PrimitiveCaster):
    def cast(self, value: Any) -> Any:
        if "allOf" in self.schema:
            for subschema in self.schema / "allOf":
                try:
                    # Note: Mutates `value` iteratively. This sequentially
                    # resolves standard overlapping types but can cause edge cases
                    # if a string is casted to an int and passed to a string schema.
                    value = self.schema_caster.evolve(subschema).cast(value)
                except (ValueError, TypeError, CastError):
                    pass

        if "oneOf" in self.schema:
            for subschema in self.schema / "oneOf":
                try:
                    # Note: Greedy resolution. Will return the first successful
                    # cast based on the order of the oneOf array.
                    return self.schema_caster.evolve(subschema).cast(value)
                except (ValueError, TypeError, CastError):
                    pass

        if "anyOf" in self.schema:
            for subschema in self.schema / "anyOf":
                try:
                    # Note: Greedy resolution. Will return the first successful
                    # cast based on the order of the anyOf array.
                    return self.schema_caster.evolve(subschema).cast(value)
                except (ValueError, TypeError, CastError):
                    pass

        return value


PrimitiveType = TypeVar("PrimitiveType")


class PrimitiveTypeCaster(Generic[PrimitiveType], PrimitiveCaster):
    primitive_type: Type[PrimitiveType] = NotImplemented

    def cast(self, value: Union[str, bytes]) -> PrimitiveType:
        return self.primitive_type(value)  # type: ignore [call-arg]


class IntegerCaster(PrimitiveTypeCaster[int]):
    primitive_type = int


class NumberCaster(PrimitiveTypeCaster[float]):
    primitive_type = float


class BooleanCaster(PrimitiveTypeCaster[bool]):
    primitive_type = bool

    def validate(self, value: Any) -> None:
        super().validate(value)

        if isinstance(value, bool):
            return

        if value.lower() not in BOOLEAN_TRUE_VALUES + BOOLEAN_FALSE_VALUES:
            raise ValueError("not a boolean format")

    def cast(self, value: Union[str, bytes]) -> bool:
        return self.primitive_type(forcebool(value))


class ArrayCaster(PrimitiveCaster):
    @property
    def items_caster(self) -> "SchemaCaster":
        # sometimes we don't have any schema i.e. free-form objects
        items_schema = self.schema.get("items", SchemaPath.from_dict({}))
        return self.schema_caster.evolve(items_schema)

    def validate(self, value: Any) -> None:
        # str and bytes are not arrays according to the OpenAPI spec
        if isinstance(value, (str, bytes)) or not isinstance(value, Iterable):
            raise ValueError("not an array format")

    def cast(self, value: list[Any]) -> list[Any]:
        return list(map(self.items_caster.cast, value))


class ObjectCaster(PrimitiveCaster):
    def validate(self, value: Any) -> None:
        if not isinstance(value, dict):
            raise ValueError("not an object format")

    def cast(self, value: dict[str, Any]) -> dict[str, Any]:
        return self._cast_proparties(value)

    def evolve(self, schema: SchemaPath) -> "ObjectCaster":
        cls = self.__class__

        return cls(
            schema,
            self.schema_validator.evolve(schema),
            self.schema_caster.evolve(schema),
        )

    def _cast_proparties(
        self, value: dict[str, Any], schema_only: bool = False
    ) -> dict[str, Any]:
        if not isinstance(value, dict):
            raise ValueError("not an object format")

        all_of_schemas = self.schema_validator.iter_all_of_schemas(value)
        for all_of_schema in all_of_schemas:
            all_of_properties = self.evolve(all_of_schema)._cast_proparties(
                value, schema_only=True
            )
            value.update(all_of_properties)

        for prop_name, prop_schema in get_properties(self.schema).items():
            try:
                prop_value = value[prop_name]
            except KeyError:
                continue
            value[prop_name] = self.schema_caster.evolve(prop_schema).cast(
                prop_value
            )

        if schema_only:
            return value

        additional_properties = self.schema.get("additionalProperties", True)
        if additional_properties is not False:
            # free-form object
            if additional_properties is True:
                additional_prop_schema = SchemaPath.from_dict(
                    {"nullable": True}
                )
            # defined schema
            else:
                additional_prop_schema = self.schema / "additionalProperties"
            additional_prop_caster = self.schema_caster.evolve(
                additional_prop_schema
            )
            for prop_name, prop_value in value.items():
                if prop_name in value:
                    continue
                value[prop_name] = additional_prop_caster.cast(prop_value)

        return value


class TypesCaster:
    casters: Mapping[str, Type[PrimitiveCaster]] = {}
    multi: Optional[Type[PrimitiveCaster]] = None

    def __init__(
        self,
        casters: Mapping[str, Type[PrimitiveCaster]],
        default: Type[PrimitiveCaster],
        multi: Optional[Type[PrimitiveCaster]] = None,
    ):
        self.casters = casters
        self.default = default
        self.multi = multi

    def get_caster(
        self,
        schema_type: Optional[Union[Iterable[str], str]],
    ) -> Type["PrimitiveCaster"]:
        if schema_type is None:
            return self.default
        if isinstance(schema_type, Iterable) and not isinstance(
            schema_type, str
        ):
            if self.multi is None:
                raise TypeError("caster does not accept multiple types")
            return self.multi

        return self.casters[schema_type]


class SchemaCaster:
    def __init__(
        self,
        schema: SchemaPath,
        schema_validator: SchemaValidator,
        types_caster: TypesCaster,
    ):
        self.schema = schema
        self.schema_validator = schema_validator

        self.types_caster = types_caster

    def cast(self, value: Any) -> Any:
        # skip casting for nullable in OpenAPI 3.0
        if value is None and (self.schema / "nullable").read_bool(
            default=False
        ):
            return value

        schema_type = (self.schema / "type").read_str(None)
        type_caster = self.get_type_caster(schema_type)

        if value is None:
            return value

        try:
            return type_caster(value)
        except (ValueError, TypeError) as exc:
            raise CastError(value, schema_type) from exc

    def get_type_caster(
        self,
        schema_type: Optional[Union[Iterable[str], str]],
    ) -> PrimitiveCaster:
        caster_cls = self.types_caster.get_caster(schema_type)
        return caster_cls(
            self.schema,
            self.schema_validator,
            self,
        )

    def evolve(self, schema: SchemaPath) -> "SchemaCaster":
        cls = self.__class__

        return cls(
            schema,
            self.schema_validator.evolve(schema),
            self.types_caster,
        )
